from contextvars import ContextVar
import json
import logging
import os
import re
import tempfile
from nemo_library.features.config import Config
from nemo_library.features.fileingestion import ReUploadFile
from nemo_library.features.focus import focusCoupleAttributes
from nemo_library.features.projects import (
    LoadReport,
    createImportedColumns,
    createOrUpdateReport,
    getImportedColumns,
    getProjectList,
)
from nemo_library.utils.migmanutils import getRelatedFields
from nemo_library.utils.utils import (
    get_display_name,
    get_import_name,
    get_internal_name,
)

__all__ = ["MigManApplyMapping"]

config_var = ContextVar("config")


def MigManApplyMapping(
    config: Config,
    mapping_fields: list[str],
    additional_fields: dict[str, str] = None,
    synonym_fields: dict[str, str] = None,
) -> None:

    # store parameters als global objects to avoid passing them to each and every funtion
    config_var.set(config)

    # build dictionary with fields to map
    
    fields_to_map = {}
    for mapping_field in mapping_fields:
        relatedfields = getRelatedFields(
            config=config,
            additionalfields=additional_fields,
            field=mapping_field,
            synonym_fields=(
                synonym_fields[mapping_field]
                if synonym_fields and mapping_field in synonym_fields
                else None
            ),
        )
        if relatedfields:
            fields_to_map[mapping_field] = relatedfields
            
    logging.info(f"mapping fields: {json.dumps(fields_to_map)}")
    return
    
    # scan all the other projects now whether they contain these fields or not
    projectList = getProjectList(config=config)["displayName"].to_list()
    dataprojects = [
        x
        for x in projectList
        if not x.startswith("Mapping")
        and not x in ["Business Processes", "Master Data"]
    ]

    for project in dataprojects:
        importedcolumns = getImportedColumns(config=config, projectname=project)[
            "displayName"
        ].to_list()

        columnstobemapped = {}
        # check for columns in this project that are mapped
        for key, values in mappingfieldsall.items():

            matched_originals = []

            # Flag to ensure ALL values in `values` must have a match
            all_match = True  # Start with the assumption that all values will match

            # Iterate over each value to check for matches in `importedcolumns`
            for val in values:
                match_found = False  # Track if the current value has a match

                for entry in importedcolumns:
                    # Check if the entry matches the pattern (value followed by a 3-digit number in parentheses)
                    if re.match(rf"^{re.escape(val)} \(\d{{3}}\)$", entry):
                        matched_originals.append(entry)  # Save the matching entry
                        match_found = True  # Mark that a match was found
                        break  # Exit the loop as we only need the first match for this value

                if not match_found:  # If no match was found for the current value
                    all_match = (
                        False  # Set the flag to False as not all values can match
                    )
                    break  # Exit the outer loop since we already know not all values match

            if all_match:  # Only proceed if ALL values in `values` matched
                columnstobemapped[key] = [
                    (src, tgt) for src, tgt in zip(values, matched_originals)
                ]

        # if there are mapped columns in this project, we are going to add mapped fields for that project now
        if columnstobemapped:

            logging.info(
                f"columns to be mapped found in project '{project}'. Here is the list: {json.dumps(columnstobemapped,indent=2)}"
            )

            # create columns for "original values" if they do not exist already
            new_columns = []
            for col in columnstobemapped:
                maplist = columnstobemapped[col]
                mapcol = f"Original_{maplist[0][1]}"
                if not mapcol in importedcolumns:
                    new_columns.append(
                        {
                            "displayName": get_display_name(mapcol),
                            "importName": get_import_name(mapcol),
                            "internalName": get_internal_name(mapcol),
                            "description": f"Original value of {maplist[0][1]}",
                            "dataType": "string",
                        }
                    )
            if new_columns:
                logging.info(f"Create Original-Columns...{json.dumps(new_columns)}")
                createImportedColumns(
                    config=config_var.get(),
                    projectname=project,
                    columns=new_columns,
                )

            sqlQuery = _SQL_Query_in_data_table(
                project=project,
                columnstobemapped=columnstobemapped,
            )

            createOrUpdateReport(
                config=config,
                projectname=project,
                displayName="(MAPPING) map data",
                querySyntax=sqlQuery,
                internalName="MAPPING_map_data",
                description="Map data",
            )
            df = LoadReport(
                config=config,
                projectname=project,
                report_name="(MAPPING) map data",
            )

            # Write to a temporary file and upload
            with tempfile.TemporaryDirectory() as temp_dir:
                temp_file_path = os.path.join(temp_dir, "tempfile.csv")

                df.to_csv(
                    temp_file_path,
                    index=False,
                    sep=";",
                    na_rep="",
                    encoding="UTF-8",
                )
                logging.info(
                    f"dummy file {temp_file_path} written for project '{project}'. Uploading data to NEMO now..."
                )

                ReUploadFile(
                    config=config,
                    projectname=project,
                    filename=temp_file_path,
                    update_project_settings=False,
                    version=3,
                    datasource_ids=[{"key": "datasource_id", "value": project}],
                )
                logging.info(f"upload to project {project} completed")

            # couple attributes
            importedcolumns = getImportedColumns(config=config, projectname=project)[
                "displayName"
            ].to_list()
            pairs = []
            for col in importedcolumns:
                original_col = f"Original_{col}"
                if original_col in importedcolumns:
                    pairs.append((col, original_col))

            for col, original_col in pairs:
                logging.info(f"Couple pairs: {col} & {original_col}")
                focusCoupleAttributes(
                    config=config,
                    projectname=project,
                    attributenames=[col, original_col],
                    previous_attribute=col,
                )


def _SQL_Query_in_data_table(
    project: str,
    columnstobemapped: dict[str, list[str]],
) -> str:

    # we start with the easy one: select all columns that are defined
    importedcolumnsdf = getImportedColumns(config=config_var.get(), projectname=project)

    # filter original-values, they will be re-created again
    importedcolumnsdf = importedcolumnsdf[
        ~importedcolumnsdf["displayName"].str.startswith("Original_")
    ]

    # add new column with display names without numbers
    importedcolumnsdf["strippedDisplayName"] = importedcolumnsdf[
        "displayName"
    ].str.replace(r" \(\d{3}\)$", "", regex=True)
    datafrags = [
        f'data.{row["internalName"]} AS "{"Original_" if row["strippedDisplayName"] in columnstobemapped.keys() else""}{row["displayName"]}"'
        for idx, row in importedcolumnsdf.iterrows()
    ]

    # now add all mapped values
    datafrags.extend(
        f'COALESCE(Mapping_{get_internal_name(key)}.TARGET_{get_internal_name(key)},data.{get_internal_name(tgt)}) AS "{tgt}"'
        for key, value in columnstobemapped.items()
        for src, tgt in value
        if src == key
    )

    # and now, we join the mapping tables
    joins = []
    for key, value in columnstobemapped.items():
        selects = [
            f"MAPPING_{get_internal_name(key)}.SOURCE_{get_internal_name(src)} = data.{get_internal_name(tgt)}"
            for src, tgt in value
        ]
        joins.append(
            f"""LEFT JOIN
    $schema.PROJECT_MAPPING_{get_internal_name(key)} MAPPING_{get_internal_name(key)}
    ON {"\n\tAND ".join(selects)}"""
        )

    # and finally build the query
    query = f"""
SELECT
    {"\n\t,".join(datafrags)}
FROM
    $schema.$table data
{"\n".join(joins)}
"""
    return query



from contextvars import ContextVar
import csv
import logging
import os
import re
from nemo_library.features.config import Config
from nemo_library.features.fileingestion import ReUploadFile
from nemo_library.features.focus import focusCoupleAttributes
from nemo_library.features.projects import (
    LoadReport,
    createImportedColumns,
    createOrUpdateReport,
    createProject,
    getImportedColumns,
    getProjectList,
)
from nemo_library.utils.migmanutils import getMappingFilePath, getRelatedFields, initializeFolderStructure
from nemo_library.utils.utils import (
    get_display_name,
    get_import_name,
    get_internal_name,
)

__all__ = ["MigManLoadMapping"]


config_var = ContextVar("config")
local_project_directory_var = ContextVar("local_project_directory")
additionalfields_var = ContextVar("additionalfields")


def MigManLoadMapping(
    config: Config,
    mapping_fields: list[str],
    local_project_directory: str,
    additionalfields: dict[str, str] = None,
    synonym_fields: dict[str, str] = None,
):

    # store parameters als global objects to avoid passing them to each and every funtion
    config_var.set(config)
    local_project_directory_var.set(local_project_directory)
    additionalfields_var.set(additionalfields)

    # if not already existing, create folder structure
    initializeFolderStructure(local_project_directory)

    # get all projects
    projectList = getProjectList(config=config)["displayName"].to_list()

    # iterate every given field and check whether to create the appropriate project and upload data
    for field in mapping_fields:

        # if project does not exist, create it
        projectname = f"Mapping {field}"
        newProject = False
        if not projectname in projectList:
            newProject = True
            createMappingProject(field=field, projectname=projectname)

            # update list of fields
            createMappingImportedColumnns(
                projectname=projectname,
                field=field,
            )
        else:
            logging.info(f"project {projectname} found.")

        # if there is data provided (as a CSV-file), we upload this now.
        # if there is no data AND the project just have been created, we upload source data and create a template file
        relatedfields = getRelatedFields(
            config=config,
            additionalfields=additionalfields,
            field=field,
            synonym_fields=(
                synonym_fields[field]
                if synonym_fields and field in synonym_fields
                else None
            ),
        )

        loadData(
            projectname=projectname,
            field=field,
            newProject=newProject,
            relatedfields=relatedfields,
        )

        # # couple attributes in focus
        coupleAttributes(projectname=projectname)

        # collect data
        collectData(
            projectname=projectname,
            field=field,
            newProject=False,  # project is not new any longer, we can access it's data (maybe uploaded in former steps)
            relatedfields=relatedfields,
        )


def coupleAttributes(
    projectname: str,
) -> None:

    imported_columns = getImportedColumns(
        config=config_var.get(),
        projectname=projectname,
    )["displayName"].to_list()
    focusCoupleAttributes(
        config=config_var.get(),
        projectname=projectname,
        attributenames=imported_columns,
        previous_attribute=None,
    )


def createMappingProject(
    projectname: str,
    field: str,
) -> str:
    """
    Creates a mapping project for a specific field if it does not already exist.

    This function checks if a project with the name "Mapping {field}" exists in the system.
    If it does not exist, it creates the project with a description. The function then
    returns the name of the project.

    Args:
        config (Config): Configuration object containing authentication and system settings.
        field (str): The name of the field for which the mapping project is to be created.

    Returns:
        str: The name of the mapping project.
    """

    logging.info(f"'{projectname}' not found, create it")
    createProject(
        config=config_var.get(),
        projectname=projectname,
        description=f"Mapping for field '{field}'",
    )


def createMappingImportedColumnns(
    projectname: str,
    field: str,
) -> dict[str, str]:

    fields = []

    additionalfields = additionalfields_var.get()
    additionalfields_filtered = additionalfields[field] if additionalfields and field in additionalfields else None
    if additionalfields_filtered:
        for additionalField in additionalfields_filtered:
            fields.append(get_display_name(f"source {additionalField}"))
    fields.append(get_display_name(f"source {field}"))
    fields.append(get_display_name(f"target {field}"))

    importedColumnsList = getImportedColumns(
        config=config_var.get(), projectname=projectname
    )
    importedColumnsList = (
        importedColumnsList["displayName"].to_list()
        if not importedColumnsList.empty
        else []
    )

    new_columns = []
    for idx, fld in enumerate(fields):
        if not fld in importedColumnsList:
            new_columns.append(
                {
                    "displayName": fld,
                    "importName": get_import_name(fld),
                    "internalName": get_internal_name(fld),
                    "description": "",
                    "dataType": "string",
                    "focusOrder": f"{idx:03}",
                }
            )
    if new_columns:
        createImportedColumns(
            config=config_var.get(),
            projectname=projectname,
            columns=new_columns,
        )


def loadData(
    projectname: str,
    field: str,
    newProject: bool,
    relatedfields: dict[str, dict[str, str]],
) -> None:

    # project is new and table does not exist. We have to upload dummy-data to enforce creation of database table

    # "real" data given? let's take this instead of the dummy file
    file_path = getMappingFilePath(projectname, local_project_directory_var.get())
    logging.info(f"checking for data file {file_path}")

    if os.path.exists(file_path):
        ReUploadFile(
            config=config_var.get(),
            projectname=projectname,
            filename=file_path,
            update_project_settings=False,
        )
        logging.info(f"upload to project {projectname} completed")
    else:
        logging.info(f"file {file_path} not found")

        if newProject:
            logging.info(
                f"file {file_path} for project {file_path} not found. Uploading source data"
            )

            queryforreport = sqlQueryInMappingTable(
                field=field,
                newProject=newProject,
                relatedfields=relatedfields,
            )

            createOrUpdateReport(
                config=config_var.get(),
                projectname=projectname,
                displayName="source mapping",
                querySyntax=queryforreport,
                description="load all source values and map them",
            )
            df = LoadReport(
                config=config_var.get(),
                projectname=projectname,
                report_name="source mapping",
            )

            # export file as a template for mappings
            df.to_csv(
                file_path,
                index=False,
                sep=";",
                na_rep="",
            )
            logging.info(f"mapping file '{file_path}' generated with source contents")

            # and upload it immediately
            ReUploadFile(
                config=config_var.get(),
                projectname=projectname,
                filename=file_path,
                update_project_settings=False,
            )
            logging.info(f"upload to project {projectname} completed")


def collectData(
    projectname: str,
    field: str,
    newProject: bool,
    relatedfields: list[str],
):

    queryforreport = sqlQueryInMappingTable(
        field=field,
        newProject=newProject,
        relatedfields=relatedfields,
    )
    createOrUpdateReport(
        config=config_var.get(),
        projectname=projectname,
        displayName="source mapping",
        querySyntax=queryforreport,
        description="load all source values and map them",
    )

    df = LoadReport(
        config=config_var.get(), projectname=projectname, report_name="source mapping"
    )

    file_path = getMappingFilePath(projectname, local_project_directory_var.get())

    # export file as a template for mappings
    df.to_csv(
        file_path,
        index=False,
        sep=";",
        na_rep="",
    )

    # and upload it immediately
    ReUploadFile(
        config=config_var.get(),
        projectname=projectname,
        filename=file_path,
        update_project_settings=False,
    )
    logging.info(f"upload to project {projectname} completed")


def sqlQueryInMappingTable(
    field: str,
    newProject: bool,
    relatedfields: list[str],
) -> str:

    # setup CTEs to load data from source projects
    ctes = []
    for ctekey, ctevalue in relatedfields.items():

        subselect = [
            f'{fldvalue} AS "{fldkey}"' for fldkey, fldvalue in ctevalue.items()
        ]

        ctes.append(
            f"""CTE_{get_internal_name(ctekey)} AS (
    SELECT DISTINCT
        {"\n\t,".join(subselect)}
    FROM 
        $schema.PROJECT_{get_internal_name(ctekey)}
)"""
        )

    # create a union for all CTEs
    globfrags = []
    for ctekey, ctevalue in relatedfields.items():

        subselect = [f'"{fldkey}"' for fldkey, fldvalue in ctevalue.items()]

        globfrags.append(
            f"""\tSELECT
    {"\n\t,".join(subselect)}
    FROM 
        CTE_{get_internal_name(ctekey)}"""
        )
    ctes.append(
        f"""CTE_ALL AS (
{"\nUNION ALL\n".join(globfrags)})"""
    )

    # and finally one for distinct value and join it with potentially existing data

    # we need to get a list of the fields itself. We assume they are the same in every CTE
    first_key = next(iter(relatedfields))
    first_value = relatedfields[first_key]
    subselect = [f'"{fldkey}"' for fldkey, fldvalue in first_value.items()]

    queryctes = f"""WITH {"\n,".join(ctes)}
,CTE_ALL_DISTINCT AS (
    SELECT DISTINCT
        {'\n\t,'.join(subselect)}
    FROM 
        CTE_ALL
)"""

    subselectsrc = [
        f'cte."{fldkey}" as "source {fldkey}"'
        for fldkey, fldvalue in first_value.items()
    ]
    subselectjoin = [
        f'mapping.SOURCE_{get_internal_name(fldkey)} = cte."{fldkey}"'
        for fldkey, fldvalue in first_value.items()
    ]
    finalquery = f"""{queryctes}
SELECT
    {'\n\t,'.join(subselectsrc)}
    , {"NULL" if newProject else f"mapping.TARGET_{get_internal_name(field)}"} AS "target {field}"
FROM
    CTE_ALL_DISTINCT cte"""

    if not newProject:
        finalquery += f"""
LEFT JOIN
    $schema.$table mapping
ON  
    {'\n\t AND '.join(subselectjoin)}"""

    return finalquery



from contextvars import ContextVar
import logging
import os
from nemo_library.features.config import Config
from nemo_library.utils.migmanutils import (
    getNEMOStepsFrompAMigrationStatusFile,
    initializeFolderStructure,
    load_database,
)
from nemo_library.utils.utils import log_error
import pandas as pd


__all__ = ["MigManCreateProjectTemplates"]

config_var = ContextVar("config")
local_project_directory_var = ContextVar("local_project_path")
projects_var = ContextVar("projects")
multi_projects_var = ContextVar("multi_projects")
database_var = ContextVar("database")


def MigManCreateProjectTemplates(
    config: Config,
    local_project_directory: str,
    projects: list[str] = None,
    proALPHA_project_status_file: str = None,
    multi_projects: dict[str, str] = None,
) -> None:

    # if there is a status file given, we ignore the given projects and get the project list from the status file
    if proALPHA_project_status_file:
        projects = getNEMOStepsFrompAMigrationStatusFile(proALPHA_project_status_file)

    # store parameters als global objects to avoid passing them to each and every funtion
    config_var.set(config)
    local_project_directory_var.set(local_project_directory)
    projects_var.set(projects)
    multi_projects_var.set(multi_projects)

    # initialize project folder structure
    initializeFolderStructure(local_project_directory)

    dbdf = load_database()
    database_var.set(dbdf)
    for project in projects:

        # check for project in database
        filtereddbdf = dbdf[dbdf["project_name"] == project]
        if filtereddbdf.empty:
            log_error(f"project '{project}' not found in database")

        # get list of postfixes
        postfixes = filtereddbdf["postfix"].unique().tolist()

        # init project
        multi_projects_list = (
            multi_projects[project] if project in multi_projects else None
        ) if multi_projects else None
        if multi_projects_list:
            for addon in multi_projects_list:
                for postfix in postfixes:
                    _create_project_template_file(project, addon, postfix)
        else:
            for postfix in postfixes:
                _create_project_template_file(project, None, postfix)


def _create_project_template_file(
    project: str,
    addon: str,
    postfix: str,
) -> None:

    logging.info(
        f"Create project template file for '{project}', addon '{addon}', postfix '{postfix}'"
    )

    dbdf = database_var.get()
    dbdf = dbdf[(dbdf["project_name"] == project) & (dbdf["postfix"] == postfix)]
    columns = dbdf["import_name"].to_list()
    data = {col: [""] for col in columns}
    templatedf = pd.DataFrame(data=data, columns=columns)
    templatedf.to_csv(
        os.path.join(
            local_project_directory_var.get(),
            "templates",
            f"{project}{" " + addon if addon else ""}{(" (" + postfix + ")") if postfix else ""}.csv",
        ),
        index=False,
        sep=";",
        encoding="UTF-8",
    )



import gzip
import logging
import shutil
import os
import time
import json
import requests
import boto3
from botocore.exceptions import NoCredentialsError
import pandas as pd

from nemo_library.features.config import Config
from nemo_library.features.projects import getProjectID
from nemo_library.utils.utils import log_error

def GetFileSize(filepath:str):
    # filesize in byte
    filesize_in_byte = os.path.getsize(filepath)
    # byte to MB (1 MB = 1024 * 1024 Byte)
    filesize_in_mb = filesize_in_byte / (1024 * 1024)
    return filesize_in_mb

def ReUploadFile(
    config: Config,
    projectname: str,
    filename: str,
    update_project_settings: bool = True,
    datasource_ids: list[dict] = None,
    global_fields_mapping: list[dict] = None,
    version: int = 2,
    trigger_only: bool = False,
) -> None:
    """
    Re-uploads a file to a specified project in the NEMO system and triggers data ingestion.

    Args:
        config (Config): Configuration object containing connection details and headers.
        projectname (str): The name of the project to which the file will be uploaded.
        filename (str): The path to the file to upload.
        update_project_settings (bool, optional): Whether to trigger the "analyze_table" task after ingestion (version 2 only). Defaults to True.
        datasource_ids (list[dict], optional): Data source identifiers for version 3 ingestion. Defaults to None.
        global_fields_mapping (list[dict], optional): Field mappings for version 3 ingestion. Defaults to None.
        version (int, optional): The ingestion version (2 or 3). Defaults to 2.
        trigger_only (bool, optional): If True, skips waiting for task completion. Defaults to False.

    Returns:
        None

    Raises:
        Exception: If any step of the file upload, data ingestion, or subsequent tasks fails.

    Notes:
        - Compresses the file into gzip format before uploading.
        - Retrieves temporary AWS S3 credentials from NEMO's Token Vendor and uploads the file to S3.
        - Sends a request to ingest the uploaded data and optionally waits for task completion.
        - Triggers "analyze_table" task if version 2 and `update_project_settings` is True.
        - Logs and raises exceptions for any errors encountered during the process.
    """

    project_id = None
    headers = None
    project_id = None
    gzipped_filename = None
    compress_level = 3
    try:
        filesize = GetFileSize(filename)

        logging.info(
            f"Size of the file: {filesize} MB"
        )

        if filesize < 5 :
            compress_level = 0
            logging.info(
                f"Compress Level: {compress_level}"
            )
            
        project_id = getProjectID(config, projectname)

        headers = config.connection_get_headers()

        logging.info(
            f"Upload of file '{filename}' into project '{projectname}' initiated..."
        )

        # Zip the file before uploading
        gzipped_filename = filename + ".gz"
        with open(filename, "rb") as f_in:
            with gzip.open(gzipped_filename, "wb", compresslevel=compress_level) as f_out:
                shutil.copyfileobj(f_in, f_out)
        logging.info(f"File {filename} has been compressed to {gzipped_filename}")

        # Retrieve temporary credentials from NEMO TVM
        response = requests.get(
            config.config_get_nemo_url()
            + "/api/nemo-tokenvendor/InternalTokenVendor/sts/s3_policy",
            headers=headers,
        )

        if response.status_code != 200:
            raise Exception(
                f"Request failed. Status: {response.status_code}, error: {response.text}"
            )

        aws_credentials = json.loads(response.text)

        aws_access_key_id = aws_credentials["accessKeyId"]
        aws_secret_access_key = aws_credentials["secretAccessKey"]
        aws_session_token = aws_credentials["sessionToken"]

        # Create an S3 client
        s3 = boto3.client(
            "s3",
            aws_access_key_id=aws_access_key_id,
            aws_secret_access_key=aws_secret_access_key,
            aws_session_token=aws_session_token,
        )

        try:
            # Upload the file
            s3filename = (
                config.config_get_tenant()
                + f"/ingestv{version}/"
                + os.path.basename(gzipped_filename)
            )
            s3.upload_file(
                gzipped_filename,
                "nemoinfrastructurestack-nemouploadbucketa98fe899-1s2ocvunlg3vs",
                s3filename,
            )
            logging.info(f"File {filename} uploaded successfully to s3 ({s3filename})")
        except FileNotFoundError:
            log_error(f"The file {filename} was not found.", FileNotFoundError)
        except NoCredentialsError:
            log_error(f"The file {filename} was not found.", NoCredentialsError)

        # Prepare data for ingestion
        data = {
            "project_id": project_id,
            "s3_filepath": f"s3://nemoinfrastructurestack-nemouploadbucketa98fe899-1s2ocvunlg3vs/{s3filename}",
        }

        if version == 3:
            if datasource_ids is not None:
                data["data_source_identifiers"] = datasource_ids
            if global_fields_mapping is not None:
                data["global_fields_mappings"] = global_fields_mapping

        endpoint_url = (
            "/api/nemo-queue/ingest_data_kubernetes_v3"
            if version == 3
            else "/api/nemo-queue/ingest_data_kubernetes_v2"
        )

        response = requests.post(
            config.config_get_nemo_url() + endpoint_url,
            headers=headers,
            json=data,
        )
        if response.status_code != 200:
            raise Exception(
                f"Request failed. Status: {response.status_code}, error: {response.text}"
            )
        logging.info("Ingestion successful")

        # Wait for task to be completed if not trigger_only
        if version == 2 or not trigger_only:
            taskid = response.text.replace('"', "")
            while True:
                data = {
                    "sort_by": "submit_at",
                    "is_sort_ascending": "False",
                    "page": 1,
                    "page_size": 20,
                }
                response = requests.get(
                    config.config_get_nemo_url() + "/api/nemo-queue/task_runs",
                    headers=headers,
                    json=data,
                )
                resultjs = json.loads(response.text)
                df = pd.json_normalize(resultjs["records"])
                df_filtered = df[df["id"] == taskid]
                if len(df_filtered) != 1:
                    raise Exception(
                        f"Data ingestion request failed, task ID not found in tasks list"
                    )
                status = df_filtered["status"].iloc[0]
                logging.info(f"Status: {status}")
                if status == "failed":
                    log_error("Data ingestion request failed, status: FAILED")
                if status == "finished":
                    if version == 2:
                        records = df_filtered["records"].iloc[0]
                        logging.info(f"Ingestion finished. {records} records loaded")
                    else:
                        logging.info("Ingestion finished.")
                    break
                time.sleep(1 if version == 2 else 5)

        # Trigger Analyze Table Task for version 2 if required
        if version == 2 and update_project_settings:
            data = {
                "project_id": project_id,
            }
            response = requests.post(
                config.config_get_nemo_url()
                + "/api/nemo-queue/analyze_table_kubernetes",
                headers=headers,
                json=data,
            )
            if response.status_code != 200:
                raise Exception(
                    f"Request failed. Status: {response.status_code}, error: {response.text}"
                )
            logging.info("Analyze_table triggered")

            # Wait for task to be completed
            taskid = response.text.replace('"', "")
            while True:
                data = {
                    "sort_by": "submit_at",
                    "is_sort_ascending": "False",
                    "page": 1,
                    "page_size": 20,
                }
                response = requests.get(
                    config.config_get_nemo_url() + "/api/nemo-queue/task_runs",
                    headers=headers,
                    json=data,
                )
                resultjs = json.loads(response.text)
                df = pd.json_normalize(resultjs["records"])
                df_filtered = df[df["id"] == taskid]
                if len(df_filtered) != 1:
                    raise Exception(
                        f"Analyze_table request failed, task ID not found in tasks list"
                    )
                status = df_filtered["status"].iloc[0]
                logging.info(f"Status: {status}")
                if status == "failed":
                    log_error("Analyze_table request failed, status: FAILED")
                if status == "finished":
                    logging.info("Analyze_table finished.")
                    break
                time.sleep(1)

    except Exception as e:
        if project_id is None:
            log_error("Upload stopped, no project_id available")
        raise log_error(f"Upload aborted: {e}")

    finally:
        if gzipped_filename:
            os.remove(gzipped_filename)



import importlib
import logging
import os
import re
import tempfile
import openpyxl
import pandas as pd

from nemo_library.features.config import Config
from nemo_library.features.fileingestion import ReUploadFile
from nemo_library.features.projects import getImportedColumns, getProjectList
from nemo_library.utils.utils import get_internal_name


def initializeFolderStructure(
    project_path: str,
) -> None:

    folders = ["templates", "mappings", "srcdata", "other"]
    for folder in folders:
        os.makedirs(os.path.join(project_path, folder), exist_ok=True)


def getMappingFilePath(projectname: str, local_project_path: str) -> str:
    return os.path.join(local_project_path, "mappings", f"{projectname}.csv")


def load_database() -> pd.DataFrame:
    with importlib.resources.open_binary(
        "nemo_library.templates", "migmantemplates.pkl"
    ) as file:
        df = pd.read_pickle(file)

    return df


def getProjectName(project: str, addon: str, postfix: str) -> str:
    return f"{project}{" " + addon if addon else ""}{(" (" + postfix + ")") if postfix else ""}"


def getNEMOStepsFrompAMigrationStatusFile(file: str) -> list[str]:
    workbook = openpyxl.load_workbook(file)
    worksheet = workbook["Status Datenübernahme"]

    data = []
    for row in worksheet.iter_rows(
        min_row=10, max_row=300, min_col=1, max_col=10, values_only=True
    ):
        data.append(row)

    # Create a DataFrame from the extracted data
    columns = [
        worksheet.cell(row=9, column=i).value for i in range(1, 11)
    ]  # Headers in row 9
    dataframe = pd.DataFrame(data, columns=columns)

    # Drop rows where "Importreihenfolge" is NaN or empty
    if "Importreihenfolge" in dataframe.columns:
        dataframe = dataframe.dropna(subset=["Importreihenfolge"])
    else:
        raise ValueError("The column 'Importreihenfolge' does not exist in the data.")

    if "Name des Importprograms / Name der Erfassungsmaske" in dataframe.columns:
        nemosteps = dataframe[dataframe["Migrationsart"] == "NEMO"][
            "Name des Importprograms / Name der Erfassungsmaske"
        ].to_list()

        nemosteps = [x.title().strip() for x in nemosteps]
        replacements = {
            "European Article Numbers": "Global Trade Item Numbers",
            "Part-Storage Areas Relationship": "Part-Storage Areas Relationships",
            "Sales Tax Id": "Sales Tax ID",
            "Mrp Parameters": "MRP Parameters",
            "Sales Units Of Measure": "Sales Units of Measure",
            "Standard Boms (Header Data)": "Standard BOMs (Header Data)",
            "Standard Boms (Line Data)": "Standard BOMs (Line Data)",
            "Routings (Standard Boms)": "Routings (Standard BOMs)",
            "Bills Of Materials For Operations (Routings Production)": "Bills of Materials for Operations (Routings Production)"
        }

        nemosteps = [
            replacements[item] if item in replacements else item for item in nemosteps
        ]

        return nemosteps
    else:
        raise ValueError(
            "The column 'Name des Importprograms / Name der Erfassungsmaske' does not exist in the data."
        )


def upload_dataframe(config: Config, project_name: str, df: pd.DataFrame):

    with tempfile.TemporaryDirectory() as temp_dir:
        temp_file_path = os.path.join(temp_dir, "tempfile.csv")

        df.to_csv(
            temp_file_path,
            index=False,
            sep=";",
            na_rep="",
            encoding="UTF-8",
        )
        logging.info(
            f"dummy file {temp_file_path} written for project '{project_name}'. Uploading data to NEMO now..."
        )

        ReUploadFile(
            config=config,
            projectname=project_name,
            filename=temp_file_path,
            update_project_settings=False,
            version=3,
            datasource_ids=[{"key": "datasource_id", "value": project_name}],
        )
        logging.info(f"upload to project {project_name} completed")

def getRelatedFields(
    config: Config,
    field: str,
    additionalfields: dict[str, str],
    synonym_fields: list[str],
) -> list[str]:
    related_fields = {}
    projectList = getProjectList(config=config)["displayName"].to_list()
    projectList = [project for project in projectList if not project.startswith("Mapping") and not project in ("Business Processes","Master Data")]
    for project in projectList:
        fields = collectDataFieldsForProject(
            config=config,
            additionalfields=additionalfields,
            project=project,
            field=field,
            synonym_fields=synonym_fields,
        )
        if fields:
            related_fields[project] = fields

    return related_fields

def collectDataFieldsForProject(
    config: Config,
    project: str,
    field: str,
    additionalfields: dict[str, str],
    synonym_fields: list[str],
) -> list[str]:

    fieldList = None
    if project in ["Business Processes", "Master Data"] or project.startswith(
        "Mapping "
    ):
        return None

    additionalfields_filtered = (
        additionalfields[field]
        if additionalfields and field in additionalfields
        else None
    )

    imported_columns = getImportedColumns(config=config, projectname=project)[
        "displayName"
    ].to_list()
    
    search_fields = [field] + synonym_fields if synonym_fields and any(synonym_fields) else [field]
    for search_field in search_fields:
        result = next(
            (
                entry
                for entry in imported_columns
                if re.match(rf"^{re.escape(search_field)} \(\d{{3}}\)$", entry)
            ),
            None,
        )
        if result:
            logging.info(f"Found field '{result}' in project '{project}'")

            fieldList = {field: get_internal_name(result)}

            # check for additional fields now
            if additionalfields_filtered:
                for additionalField in additionalfields_filtered:
                    result = next(
                        (
                            entry
                            for entry in imported_columns
                            if re.match(
                                rf"^{re.escape(additionalField)} \(\d{{3}}\)$", entry
                            )
                        ),
                        None,
                    )
                    if not result:
                        logging.info(
                            f"Field '{additionalField}' not found in project '{project}'. Skip this project"
                        )

                    fieldList[additionalField] = get_internal_name(result)

    # we have found all relevant fields in project. Now we are going to collect data
    return fieldList
